@@ -1,6 +1,6 @@
 //! Provides high-level access to the redstone APIs.
 
-use crate::common::{Colour, Side, BLOCK_SIDES, COLOURS};
+use crate::common::{Colour, Lockable, Side, BLOCK_SIDES, COLOURS};
 use crate::error::Error;
 use crate::helpers::{Ignore, OneValue, ThreeValues, TwoValues};
 use alloc::vec::Vec;
@@ -31,17 +31,12 @@ impl Redstone {
 	pub fn address(&self) -> &Address {
 		&self.0
 	}
+}
 
-	/// Locks the component so methods can be invoked on it.
-	///
-	/// The [`Invoker`](Invoker) and a scratch buffer must be provided. They are released and can
-	/// be reused once the [`Locked`](Locked) is dropped.
-	#[must_use = "This function is only useful for its return value"]
-	pub fn lock<'invoker, 'buffer>(
-		&self,
-		invoker: &'invoker mut Invoker,
-		buffer: &'buffer mut Vec<u8>,
-	) -> Locked<'invoker, 'buffer> {
+impl<'invoker, 'buffer> Lockable<'invoker, 'buffer> for Redstone {
+	type Locked = Locked<'invoker, 'buffer>;
+
+	fn lock(&self, invoker: &'invoker mut Invoker, buffer: &'buffer mut Vec<u8>) -> Self::Locked {
 		Locked {
 			address: self.0,
 			invoker,
@@ -79,8 +74,7 @@ impl<'invoker, 'buffer> Locked<'invoker, 'buffer> {
 	/// indices should be [relative sides](crate::common::RelativeSide).
 	///
 	/// # Errors
-	/// * [`BadComponent`](Error::BadComponent) is returned if the card or block does not exist, is
-	///   inaccessible, or is not a redstone component.
+	/// * [`BadComponent`](Error::BadComponent)
 	pub async fn get_input(&mut self) -> Result<[u8; BLOCK_SIDES], Error> {
 		self.get_vanilla("getInput").await
 	}
@@ -92,8 +86,7 @@ impl<'invoker, 'buffer> Locked<'invoker, 'buffer> {
 	/// must be a [relative side](crate::common::RelativeSide).
 	///
 	/// # Errors
-	/// * [`BadComponent`](Error::BadComponent) is returned if the card or block does not exist, is
-	///   inaccessible, or is not a redstone component.
+	/// * [`BadComponent`](Error::BadComponent)
 	pub async fn get_side_input(&mut self, side: impl Side) -> Result<u8, Error> {
 		self.get_side_vanilla("getInput", side.into()).await
 	}
@@ -106,8 +99,7 @@ impl<'invoker, 'buffer> Locked<'invoker, 'buffer> {
 	/// array is itself an array indexed by colour.
 	///
 	/// # Errors
-	/// * [`BadComponent`](Error::BadComponent) is returned if the card or block does not exist, is
-	///   inaccessible, or is not a redstone component.
+	/// * [`BadComponent`](Error::BadComponent)
 	pub async fn get_bundled_input(&mut self) -> Result<[[u8; COLOURS]; BLOCK_SIDES], Error> {
 		self.get_bundled("getBundledInput").await
 	}
@@ -121,8 +113,7 @@ impl<'invoker, 'buffer> Locked<'invoker, 'buffer> {
 	/// The returned array is indexed by colour.
 	///
 	/// # Errors
-	/// * [`BadComponent`](Error::BadComponent) is returned if the card or block does not exist, is
-	///   inaccessible, or is not a redstone component.
+	/// * [`BadComponent`](Error::BadComponent)
 	pub async fn get_side_bundled_input(
 		&mut self,
 		side: impl Side,
@@ -137,8 +128,7 @@ impl<'invoker, 'buffer> Locked<'invoker, 'buffer> {
 	/// must be a [relative side](crate::common::RelativeSide).
 	///
 	/// # Errors
-	/// * [`BadComponent`](Error::BadComponent) is returned if the card or block does not exist, is
-	///   inaccessible, or is not a redstone component.
+	/// * [`BadComponent`](Error::BadComponent)
 	pub async fn get_side_colour_bundled_input(
 		&mut self,
 		side: impl Side,
@@ -155,8 +145,7 @@ impl<'invoker, 'buffer> Locked<'invoker, 'buffer> {
 	/// indices should be [relative sides](crate::common::RelativeSide).
 	///
 	/// # Errors
-	/// * [`BadComponent`](Error::BadComponent) is returned if the card or block does not exist, is
-	///   inaccessible, or is not a redstone component.
+	/// * [`BadComponent`](Error::BadComponent)
 	pub async fn get_output(&mut self) -> Result<[u8; BLOCK_SIDES], Error> {
 		self.get_vanilla("getOutput").await
 	}
@@ -168,8 +157,7 @@ impl<'invoker, 'buffer> Locked<'invoker, 'buffer> {
 	/// must be a [relative side](crate::common::RelativeSide).
 	///
 	/// # Errors
-	/// * [`BadComponent`](Error::BadComponent) is returned if the card or block does not exist, is
-	///   inaccessible, or is not a redstone component.
+	/// * [`BadComponent`](Error::BadComponent)
 	pub async fn get_side_output(&mut self, side: impl Side) -> Result<u8, Error> {
 		self.get_side_vanilla("getOutput", side.into()).await
 	}
@@ -182,8 +170,7 @@ impl<'invoker, 'buffer> Locked<'invoker, 'buffer> {
 	/// array is itself an array indexed by colour.
 	///
 	/// # Errors
-	/// * [`BadComponent`](Error::BadComponent) is returned if the card or block does not exist, is
-	///   inaccessible, or is not a redstone component.
+	/// * [`BadComponent`](Error::BadComponent)
 	pub async fn get_bundled_output(&mut self) -> Result<[[u8; COLOURS]; BLOCK_SIDES], Error> {
 		self.get_bundled("getBundledOutput").await
 	}
@@ -197,8 +184,7 @@ impl<'invoker, 'buffer> Locked<'invoker, 'buffer> {
 	/// The returned array is indexed by colour.
 	///
 	/// # Errors
-	/// * [`BadComponent`](Error::BadComponent) is returned if the card or block does not exist, is
-	///   inaccessible, or is not a redstone component.
+	/// * [`BadComponent`](Error::BadComponent)
 	pub async fn get_side_bundled_output(
 		&mut self,
 		side: impl Side,
@@ -213,8 +199,7 @@ impl<'invoker, 'buffer> Locked<'invoker, 'buffer> {
 	/// must be a [relative side](crate::common::RelativeSide).
 	///
 	/// # Errors
-	/// * [`BadComponent`](Error::BadComponent) is returned if the card or block does not exist, is
-	///   inaccessible, or is not a redstone component.
+	/// * [`BadComponent`](Error::BadComponent)
 	pub async fn get_side_colour_bundled_output(
 		&mut self,
 		side: impl Side,
@@ -235,8 +220,7 @@ impl<'invoker, 'buffer> Locked<'invoker, 'buffer> {
 	/// The old signal levels, prior to modification, are returned.
 	///
 	/// # Errors
-	/// * [`BadComponent`](Error::BadComponent) is returned if the card or block does not exist, is
-	///   inaccessible, or is not a redstone component.
+	/// * [`BadComponent`](Error::BadComponent)
 	pub async fn set_output(
 		&mut self,
 		levels: &[Option<u8>; BLOCK_SIDES],
@@ -261,8 +245,7 @@ impl<'invoker, 'buffer> Locked<'invoker, 'buffer> {
 	/// The old signal level, prior to modification, is returned.
 	///
 	/// # Errors
-	/// * [`BadComponent`](Error::BadComponent) is returned if the card or block does not exist, is
-	///   inaccessible, or is not a redstone component.
+	/// * [`BadComponent`](Error::BadComponent)
 	pub async fn set_side_output(&mut self, side: impl Side, level: u8) -> Result<u8, Error> {
 		let side: u8 = side.into();
 		let ret: OneValue<_> = component_method(
@@ -286,8 +269,7 @@ impl<'invoker, 'buffer> Locked<'invoker, 'buffer> {
 	/// colour unmodified.
 	///
 	/// # Errors
-	/// * [`BadComponent`](Error::BadComponent) is returned if the card or block does not exist, is
-	///   inaccessible, or is not a redstone component.
+	/// * [`BadComponent`](Error::BadComponent)
 	pub async fn set_bundled_output(
 		&mut self,
 		levels: &[[Option<u8>; COLOURS]; BLOCK_SIDES],
@@ -322,8 +304,7 @@ impl<'invoker, 'buffer> Locked<'invoker, 'buffer> {
 	/// `Some` with the new signal strength to emit, or `None` to leave that colour unmodified.
 	///
 	/// # Errors
-	/// * [`BadComponent`](Error::BadComponent) is returned if the card or block does not exist, is
-	///   inaccessible, or is not a redstone component.
+	/// * [`BadComponent`](Error::BadComponent)
 	pub async fn set_side_bundled_output(
 		&mut self,
 		side: impl Side,
@@ -350,8 +331,7 @@ impl<'invoker, 'buffer> Locked<'invoker, 'buffer> {
 	/// The old signal level, prior to modification, is returned.
 	///
 	/// # Errors
-	/// * [`BadComponent`](Error::BadComponent) is returned if the card or block does not exist, is
-	///   inaccessible, or is not a redstone component.
+	/// * [`BadComponent`](Error::BadComponent)
 	pub async fn set_side_colour_bundled_output(
 		&mut self,
 		side: impl Side,
@@ -381,8 +361,7 @@ impl<'invoker, 'buffer> Locked<'invoker, 'buffer> {
 	/// adjacent block. If the target block is not readable by a comparator, zero is returned.
 	///
 	/// # Errors
-	/// * [`BadComponent`](Error::BadComponent) is returned if the card or block does not exist, is
-	///   inaccessible, or is not a redstone component.
+	/// * [`BadComponent`](Error::BadComponent)
 	pub async fn get_comparator_input(&mut self, side: impl Side) -> Result<u8, Error> {
 		let side: u8 = side.into();
 		let ret: OneValue<_> = component_method(
@@ -403,8 +382,7 @@ impl<'invoker, 'buffer> Locked<'invoker, 'buffer> {
 	/// connected computers (in the case of a redstone block) are powered on if they are off.
 	///
 	/// # Errors
-	/// * [`BadComponent`](Error::BadComponent) is returned if the card or block does not exist, is
-	///   inaccessible, or is not a redstone component.
+	/// * [`BadComponent`](Error::BadComponent)
 	pub async fn get_wake_threshold(&mut self) -> Result<u32, Error> {
 		let ret: OneValue<_> = component_method::<(), _>(
 			self.invoker,
@@ -426,8 +404,7 @@ impl<'invoker, 'buffer> Locked<'invoker, 'buffer> {
 	/// The old wake threshold, prior to modification, is returned.
 	///
 	/// # Errors
-	/// * [`BadComponent`](Error::BadComponent) is returned if the card or block does not exist, is
-	///   inaccessible, or is not a redstone component.
+	/// * [`BadComponent`](Error::BadComponent)
 	pub async fn set_wake_threshold(&mut self, threshold: u32) -> Result<u32, Error> {
 		let ret: OneValue<_> = component_method(
 			self.invoker,
@@ -509,7 +486,7 @@ impl<'invoker, 'buffer> Locked<'invoker, 'buffer> {
 /// [`Default`](Default), and any elements not included in the CBOR-encoded map will be returned at
 /// their default values.
 ///
-/// To be decoded, the element type must be an [`Option`](Option) and the type contained therein
+/// To be encoded, the element type must be an [`Option`](Option) and the type contained therein
 /// must implement [`Encode`](minicbor::Encode). `None` values are omitted from the map entirely,
 /// while `Some` values are encoded as their contents, keyed by array position.
 #[derive(Clone, Copy)]
@@ -522,7 +499,9 @@ impl<'buffer, T: Copy + Decode<'buffer> + Default, const LENGTH: usize> Decode<'
 		let mut ret = [T::default(); LENGTH];
 		// The CBOR fits in memory, so it must be <2³² elements.
 		#[allow(clippy::cast_possible_truncation)]
-		let length = d.map()?.ok_or(minicbor::decode::Error::Message(""))? as usize;
+		let length = d
+			.map()?
+			.ok_or_else(|| minicbor::decode::Error::message(""))? as usize;
 		for _ in 0..length {
 			let key = d.u32()?;
 			ret[key as usize] = d.decode::<T>()?;
